//
//  BaseSA.cpp
//
//  2004/07/26
//  2004/10/18




#include  <math.h>
#include  <stdlib.h>
#include  <stdio.h>

#include  "../Random.h"
#include  "../Tour.h"
#include  "../TSP.h"

double  getInitTemp( Tour* sa );



//  command-line arguments
//    0  -  command of invocation
//    1  -  number of "threads"
//    2  -  input file
//    3  -  output file  (not used as originally)
//    4  -  mu
//    5  -  [for SAGA only]
//    6  -  srand() seed

//  Here are the default command-line arguments:

char*  args[] = { "BaseSA",
  "1", 
  "C:\\q\\dev\\yi\\problemFile.txt",
  "C:\\q\\dev\\yi\\_default.out.txt",
  "0.9",		
  "0.9",
  "1"
};


const unsigned  maxTemperatureSteps = 100u;
const unsigned  maxIterations = 2000000u;		// 20041017

TSP*  pTSP;		// which TSP to solve
double  mu;		// cooling parameter



int  main( unsigned argc, char* argv[] )
{

	FILE*  writer;			// output file
	unsigned  bestLength=0xffffffffu;

	double  currentTemp;
	double  deltaE;

	unsigned  i;
	unsigned  j;

// Copy command-line values from argv to args here.
	for ( unsigned a=1u; a<argc; a++ )
		args[a]  =  argv[a];

	writer  =  fopen( args[3], "w" );
	if  ( !writer )
	{
		fprintf( stderr, "*** BaseSA: unable to open output file ***\n" );
		fprintf( stderr, "      %s\n", argv[3] );
	}

	mu  =  atof(args[4]);		// Set the cooling parameter.
	srand( atoi(args[6]) );		// Set the rand() seed.



	pTSP  =  new TSP(args[2]);
	Tour*  sa  =  new Tour(pTSP);

	sa->setBest();

	bestLength  =  sa->getCurrentLength();
	printf( "bestLength = %d\n", bestLength );


	//  Use the adaptive initial temperature setter.
	currentTemp  =  getInitTemp( sa );


	for ( i=0u; i < maxTemperatureSteps; i++ )
	{
		//  inner loop
		for ( j=0u; j < maxIterations; j++ )
		{
			sa->twoOpt();
			deltaE  =  double(sa->getImprovement());

			if ( deltaE >= 0. )
				sa->update();
			else
				if (Random::random() < exp(deltaE/currentTemp))
					sa->update();
		}
		//  end of inner loop

		sa->backToBest();

		if ( sa->getCurrentLength() < bestLength )
		{
			currentTemp  =  currentTemp * mu;
			bestLength  =  sa->getCurrentLength();
		}

		printf( "%3d", i );
		printf( ",   currentTemp[%1d] = %g", 0, currentTemp );
		printf( ",   %10d\n", sa->getCurrentLength());
	}
	//  end of outer loop


	fprintf( writer, "%d\n", sa->getCurrentLength() );

	fclose( writer );		// good housekeeping
	delete  sa;
	delete  pTSP;
	return  0;
}
