//
//  TSP.cpp
//  2004/08/07
//  2004/10/16
//


#include  "TSP.h"


#include  <math.h>
#include  <stdio.h>



TSP::TSP( char* file )
	:  N(0u), xCoord((int*)0xffffffff), yCoord((int*)0xffffffff),
	distances((double*)0xffffffff),
	filename((char*)0xffffffff)
{
	filename  =  file;		//  Save filename for miscellaneous operations.

	N  =  readN( file );

	xCoord  =  new int[N];
	yCoord  =  new int[N];

	loadCoords( file, N, xCoord, yCoord );
}




unsigned  TSP::getLength( int* array )
{
	unsigned  total = 0u;

	for (unsigned i=1; i<N; i++)
		total += getDistance(array[i-1], array[i]);
	total += getDistance(array[N-1], array[0]);

	return total;
}





void  TSP::makeDistances()
{
	double  xDist, yDist;
	double  dist=0.0;

	for ( unsigned i=0; i<N; i++)
	{
		for ( unsigned j=0; j<N; j++)
		{
			xDist = xCoord[i] - xCoord[j];
			yDist = yCoord[i] - yCoord[j];
			dist = sqrt(xDist*xDist + yDist*yDist);

			distances[i*N+j] = (int) floor(dist + 0.5);
		}
	}
}




void  TSP::loadCoords( char* file, unsigned n, int* xCoord, int* yCoord )
{
	int  skippee;
	double  inVal;

	FILE*  reader  =  fopen( file , "r" );
	if  ( !reader )
	{
		fprintf( stderr, "*** TSP::loadCoords(): unable to open input file ***\n" );
		fprintf( stderr, "%s\n", file );
		fflush( stderr );
	}

	fscanf( reader, "%d", &skippee );

	for ( unsigned i=0; i<n; i++ )
	{
		fscanf( reader, "%d", &skippee );
		fscanf( reader, "%lf", &inVal );
		xCoord[i] = (int) inVal;
		fscanf( reader, "%lf", &inVal );
		yCoord[i] = (int) inVal;
	}
	fclose(reader);
}



unsigned  TSP::readN( char* file )
{
	unsigned  retVal = 0xffffffffu;

	FILE*  reader  =  fopen( file , "r" );
	fscanf( reader, "%d", &retVal );
	fclose( reader );

	return  retVal;
}



bool  TSP::coordCheck()
{
	bool  retVal = true;

	int* xRef  =  new int[N];
	int* yRef  =  new int[N];
	loadCoords( filename, N, xRef, yRef );

	for ( unsigned k=0; k<N; k++)
		if ( xCoord[k] != xRef[k]  ||  yCoord[k] != yRef[k] )
			retVal  =  false;

	delete[]  xRef;		// housekeeping
	delete[]  yRef;

	return  retVal;
}
