//
//  Tour.cpp
//
//  2004/08/07
//  2004/10/30


#include  "Tour.h"
#include  "MyStack.h"
#include  "Random.h"
#include  "TSP.h"


#include  <math.h>
#include  <stdio.h>
#include  <stdlib.h>
#include  <string.h>



Tour::Tour()
	:  N(0xffffffff), edge1(0xffffffff), edge2(0xffffffff), improvement(0xffffffff),
		bestLength(0xffffffff), currentLength(0xffffffff),
		best(NULL), current(NULL), random(NULL),
		forwardEdge(NULL), backwardEdge(NULL), 
		tspInstance(NULL)
	// (some of the above pointers subject to a delete[] attempt)
{


///  Note that 8192 was hard-coded to simplify a thread synchronization issue.
	best  =  new int[8192];			///***
	current  =  new int[8192];		///***
	random  =  new int[8192];		///***

	forwardEdge  =  new int[8192];	///***
	backwardEdge  =  new int[8192];	///***
///  Must be changed if optimizing for more than 8192 cities.

}




Tour::Tour( TSP* instance )
	:  N(0u), edge1(0u), edge2(0u), improvement(0u),
		bestLength(0xffffffff), currentLength(0xffffffff),
		best((int*)0xffffffff), current((int*)0xffffffff),
		random((int*)0xffffffff)
{
	unsigned  i=0u;

	N  = instance->getN();
	tspInstance  = instance;

	best  =  new int[N];
	current  =  new int[N];
	random  =  new int[N];

	makeRandom();


	pStack  =  (MyStack*)0xffffffff;
	pStack  =  new MyStack(N);

	forwardEdge  =  (int*)0xffffffff;
	forwardEdge  =  new int[N];
	backwardEdge  =  (int*)0xffffffff;
	backwardEdge  =  new int[N];
}



Tour::~Tour()
{
	delete[]  best;
	delete[]  current;
	delete[]  random;

	delete[]  forwardEdge;
	delete[]  backwardEdge;

	delete  pStack;
}



Tour&  Tour::operator=( const Tour& rhs )
{
	unsigned  i=0u;

	N  =  rhs.N;
	tspInstance  = rhs.tspInstance;

	bestLength  =  rhs.bestLength;
	currentLength  =  rhs.currentLength;
	improvement  =  rhs.improvement;
	edge1  =  rhs.edge1;
	edge2  =  rhs.edge2;

	delete[]  best;
	delete[]  current;
	delete[]  random;
	delete[]  forwardEdge;
	delete[]  backwardEdge;

	best  =  new int[N];
	current  =  new int[N];
	random  =  new int[N];

	forwardEdge  =  new int[N];
	backwardEdge  =  new int[N];

	for  ( i=0; i<N; i++)		///
	{							///
		best[i]  =  rhs.best[i];		///
		current[i]  =  rhs.current[i];	///
		random[i]  =  rhs.random[i];		///

		forwardEdge[i]  =  rhs.forwardEdge[i];
		backwardEdge[i]  =  rhs.backwardEdge[i];
	}

	delete  pStack;
	pStack  =  (MyStack*)0xffffffff;
	pStack  =  new MyStack(N);

	return  *this;
}



void  Tour::backToBest()
{
	backToBest(false);
}



void  Tour::backToBest( const bool& make )
{
	memcpy( current, best, N*sizeof(int) );
	currentLength  =  bestLength;

	if (make)
		makeEdges();
}



void  Tour::calculateImprovement()
{
	int  left1  =  current[(edge1+(N-1))%N];
	int  left2  =  current[edge2-1];
	int  right1  =  current[edge1];
	int  right2  =  current[edge2];

	unsigned  currentDist  =  tspInstance->getDistance(left1, right1)
		+ tspInstance->getDistance(left2, right2);

	unsigned  newDist  =  tspInstance->getDistance(left1, left2)
		+ tspInstance->getDistance(right1, right2);

	improvement  =  currentDist - newDist;
}




void  Tour::copyTour( Tour* from )
{
	unsigned  i=0u;

	N  =  from->N;
	tspInstance  =  from->tspInstance;

	bestLength  =  from->bestLength;
	currentLength  =  from->currentLength;
	improvement  =  from->improvement;
	edge1  =  from->edge1;
	edge2  =  from->edge2;

	for  ( i=0; i<N; i++)		///
	{							///
		best[i]  =  from->best[i];			///
		current[i]  =  from->current[i];		///
		random[i]  =  from->random[i];		///

		forwardEdge[i]  =  from->forwardEdge[i];
		backwardEdge[i]  =  from->backwardEdge[i];
	}

}




int  Tour::extraDistance( const unsigned& first, const unsigned& second, const unsigned& insert)
{
	int  diff = 0;
	int  present  =  tspInstance->getDistance(first, second);
	int  extra  =  tspInstance->getDistance(first, insert) +
		           tspInstance->getDistance(insert, second);
		
	diff  =  extra - present;

	return  diff;
}



int  Tour::findDifferent( int* other )
{
	int  retVal=0;

	int  base = (int) floor(N*Random::random());
	unsigned  rIndex = 0u;
	int  first=0;
	int  second=0;

	retVal  =  (int) floor(N*Random::random());
	for ( unsigned i=0u; i<N; i++ )
	{
		rIndex  =  random[(base+i)%N];
		second  =  other[rIndex];
		first  =  other[(rIndex+(N-1))%N];
		if (forwardEdge[second] != first && backwardEdge[second] != first)
		{
			retVal  =  rIndex;
			break;
		}
	}

    return  retVal;
}



void  Tour::loadTour( char* file )
{
	int  inputVal=0;
	FILE*  reader  =  fopen( file, "r" );

	fscanf( reader, "%d", &inputVal );		// read & discard

	for ( unsigned i=0; i<N; i++ )
	{

		fscanf( reader, "%d", &inputVal );	///
		current[i]  =  inputVal - 1;		///
	}

	fclose(reader);

	currentLength  =  tspInstance->getLength(current);	//
}



void  Tour::makeEdges ()
{
	unsigned  i=0u;
	for ( i=1; (i+1)<N; i++ )
	{
		forwardEdge[current[i]] = current[i+1];
		backwardEdge[current[i]] = current[i-1];
	}
	forwardEdge[current[0]] = current[1];
	forwardEdge[current[N-1]] = current[0];
	backwardEdge[current[0]] = current[N-1];
	backwardEdge[current[N-1]] = current[N-2];
}



void Tour::makeRandom ()
{
	unsigned  i=0u;

	for ( i=0; i < N; i++ )
		current[i] = i;

	unsigned  target=0u;		///
	for ( i=0; i < N; i++ )
	{
		target = (int) (Random::random() * (N-i));
		swap(current, i, target);
	}

	currentLength  =  tspInstance->getLength(current);	//

	memcpy( random, current, N*sizeof(int) );		///
	memcpy( best, current, N*sizeof(int) );			///
	bestLength  =  currentLength;
}



void  Tour::printTour()
{
	printf( "Valid tour = %d\n", validate() );
	printf( "Length = %f\n", currentLength );
}



void  Tour::printTour(char* file)
{
	FILE*  writer = fopen( file , "w" );

	fprintf( writer, "Valid tour = %d\n", validate() );
	fprintf( writer, "Length = %d\n", currentLength );
	fprintf( writer, "Cities\n");
	for ( unsigned i=0; i<N; i++ )
	{
		fprintf( writer, "%d\n", current[i] );
	}
	fclose(writer);
}



void  Tour::setBest()
{
	setBest(false);
}



void  Tour::setBest(const bool& make)
{
	memcpy( best, current, N*sizeof(int) );		///
	bestLength  =  currentLength;
	if (make)
		makeEdges();
}



void  Tour::swap( int* array, const unsigned& first, const unsigned& second )
{
	int temp  =  array[first];
	array[first]  =  array[second];
	array[second]  =  temp;
}



void  Tour::twoOpt()
{
	twoOpt( NULL, 0.0 );
}



void  Tour::twoOpt( int* array )
{
	unsigned  index=0u;
	for ( index=edge1; index < edge2; index++ )
		pStack->push(array[index]);

	for ( index=edge1; index < edge2; index++ )
		array[index]  =  pStack->pop();
}



void  Tour::twoOpt( Tour* other, const double& saga )
{
	unsigned  max=0u;
	unsigned  edges=0u;

	double neighbourhood  =  Random::random();
	max  =  int(N/(neighbourhood*50+2));

	if (other == NULL)
	{
		edge1 = (int) floor(N*Random::random());
	}
	else
	{
		if (Random::random() < saga)
			edge1 = other->findDifferent(current);
		else
			edge1 = (int) floor(N*Random::random());
	}

	edges  =  2 + (int) floor(max*Random::random());
	edge2  =  edge1 + edges;
	if ( edge2 >= N )
	{
		edge2  =  edge1;
		edge1  =  (edge1 + edges)%N;
	}
	calculateImprovement();
}



void  Tour::update()
{
	twoOpt(current);

	currentLength  -=  improvement;		// <<<<<<

	if (currentLength < bestLength)
	{
		bestLength  =  currentLength;
		memcpy( best, current, N*sizeof(int) );		///
	}
}


/////////////////////////////////////////////////////


bool  Tour::validate()
{
	bool  retVal  =  true;
	unsigned  i=0u;

	currentLength  =  tspInstance->getLength(current);	// <<<<<<

	bool* visited  =  new bool[N];
	for ( i=0; i < N; i++ )
		visited[i] = false;

	for ( i=0; i < N; i++ )
		visited[current[i]] = true;

	for ( i = 0; i < N; i++ )
		if (!visited[i])
			retVal  =  false;

	delete[]  visited;		///

	return retVal;
}



bool  Tour::listCheck( int* list )
{
	bool  retVal = true;
	unsigned  i, j;
	unsigned*  count;

	count  =  new unsigned[N];

	for ( i=0; i < N; i++ )
		count[i]  =  0;

	for ( i=0; i < N; i++ )
	{
		j  =  list[i];
		if  ( 0 <= j  &&  j < N )
			count[j]++;
	}

	for ( i=0; i < N; i++ )
	{
		if  ( count[i] != 1 )
		{
			retVal  =  false;
			break;
		}
	}

	delete[]  count;
	return  retVal;
}
