//
//  thread1.cpp  -  for use with mtSA
//
//  2004/10/24
//


#include  <windows.h>

#include  <math.h>
#include  <stdlib.h>
#include  <stdio.h>

#include  "../Random.h"
#include  "../Tour.h"
#include  "../TSP.h"

double  getInitTemp( Tour* sa );

inline void  copySolution( Tour* to, Tour* from );
inline unsigned  getRandomPartner( const unsigned& butNotThis );
inline unsigned  getSolLength( const unsigned& thrIndex );
inline bool  isPartnerSolutionBetter( const unsigned& self, const unsigned& iPartner );
inline void  postSolution( const unsigned& threadParam, Tour* sa );




const unsigned  maxTemperatureSteps = 100u;
const unsigned  maxIterations = 2000000u;

extern unsigned  nThreads;	// number of threads

extern double  mu;		// cooling parameter
extern TSP*  pTSP;
extern char*  outFileNameBase;

//  thread synchronization events
extern HANDLE  hEvent3[];	// signal thread completion

extern CRITICAL_SECTION  cs;		// critical section for copying solutions

//  results which can be shared across threads
extern Tour  postedSolution[];
extern unsigned  sLength[];





DWORD WINAPI  threadCode( LPVOID threadParam )
{
	unsigned  self = unsigned(threadParam);

	unsigned  bestLength = ~0u;	// largest integer

	double  currentTemp;
	double  deltaE;

	unsigned  i;
	unsigned  j;

	unsigned  iPartner;		// thread index of solution-sharing partner


	Tour*  sa  =  new Tour(pTSP);

	sa->setBest();

	bestLength  =  sa->getCurrentLength();
	printf( "bestLength = %d\n", bestLength );


	//  Use the adaptive initial temperature setter.
	currentTemp  =  getInitTemp( sa );

	//  Information which can be shared with other threads:
	postSolution( self, sa );



	//  Indicate that this thread has the temperature set,
	//    and wait for other threads to set theirs.


	//  outer loop
	for ( i=0u; i < maxTemperatureSteps; i++ )
	{

		//  inner loop
		for ( j=0u; j < maxIterations; j++ )
		{

			sa->twoOpt();
			deltaE  =  double(sa->getImprovement());

			if ( deltaE >= 0. )
				sa->update();
			else
				if (Random::random() < exp(deltaE/currentTemp))
					sa->update();
		}
		//  end of inner loop

		sa->backToBest(true);

		if ( sa->getCurrentLength() < bestLength )
		{
			currentTemp  =  currentTemp * mu;
			bestLength  =  sa->getCurrentLength();
		}


		printf( "%3d:%3d", self, i );
		printf( ",   currentTemp[%1d] = %g", 0, currentTemp );
		printf( ",   %10d\n", sa->getCurrentLength());


		postSolution( self, sa );



//  communication with other threads

		if ( nThreads > 1u )
		{
			iPartner  =  getRandomPartner( self );

			if ( isPartnerSolutionBetter(self,iPartner) )
			{
				// copy solution.
				copySolution( sa, postedSolution+iPartner );
			}
		}
		else  //  i.e., just one thread
		{
			postedSolution[0]  =  *sa;
		}

	}
	//  end of outer loop


	// reporting

	sLength[self]  =  sa->getCurrentLength();


	delete  sa;			// good housekeeping

	return  ! SetEvent( hEvent3[self] );
}






//  protected copy between threads
void  copySolution( Tour* to, Tour* from )
{
	EnterCriticalSection( &cs );
	*to  =  *from;
	LeaveCriticalSection( &cs );
}




unsigned  getRandomPartner( const unsigned& butNotSelf )
{
	unsigned  retVal = unsigned(Random::random()*nThreads);

	while ( retVal == butNotSelf )
		retVal  =  unsigned(Random::random()*nThreads);

	return  retVal;
}




unsigned  getSolLength( const unsigned& thrIndex )
{
	return  postedSolution[thrIndex].getCurrentLength();
}




bool  isPartnerSolutionBetter( const unsigned& self, const unsigned& iPartner )
{
	bool  retVal;
	retVal  =  getSolLength(iPartner) < getSolLength(self);
	return  retVal;
}




void  postSolution( const unsigned& self, Tour* sa )
{
	copySolution( postedSolution+self, sa );
}
