//
//  thread2.cpp  --  for use with mtSAGA
//
//  2004/10/16
//  2004/10/29


#include  <windows.h>

#include  <math.h>
#include  <stdlib.h>
#include  <stdio.h>

#include  "../Random.h"
#include  "../Tour.h"
#include  "../TSP.h"

double  getInitTemp( Tour* sa );

inline void  copySolution( Tour* to, Tour* from );
inline unsigned  getIndexOfLatest( const unsigned& k );
inline unsigned  getRandomPartner( const unsigned& butNotThis );
inline unsigned  getSolLength( const unsigned& thrIndex );
inline void  postSolution( const unsigned& threadParam, Tour* sa );




const unsigned  maxTemperatureSteps = 100u;
const unsigned  maxIterations = 2000000u;

extern unsigned  nThreads;	// number of threads

extern double  mu;		// cooling parameter
extern double  common;	// common components parameter
extern TSP*  pTSP;
extern char*  outFileNameBase;

//  thread synchronization events
extern HANDLE  hEvent1[];	// signal temperature-set completion
extern HANDLE  hEvent2;		// signal thread resumption
extern HANDLE  hEvent3[];	// signal thread completion

extern CRITICAL_SECTION  cs;		// critical section for copying solutions

//  results from the threads
extern Tour  postedSolution[];
extern unsigned  sLength[];




DWORD WINAPI  threadCode( LPVOID threadParam )
{
	unsigned  self = unsigned(threadParam);

	unsigned  bestLength = ~0u;	// largest integer

	double  currentTemp;
	double  deltaE;

	unsigned  i;
	unsigned  j;

	unsigned  iPartner;		// thread index of solution-sharing partner


	Tour*  sa  =  new Tour(pTSP);
	Tour*  pPsa  =  new Tour(pTSP);

	sa->setBest();

	bestLength  =  sa->getCurrentLength();
	printf( "bestLength = %d\n", bestLength );


	//  Use the adaptive initial temperature setter.
	currentTemp  =  getInitTemp( sa );


	//  Information which can be shared with other threads:
	postSolution( self, sa );



	//  Indicate that this thread has the temperature set,
	//    and wait for other threads to set theirs.
	SetEvent( hEvent1[self] );
	WaitForSingleObject( hEvent2, INFINITE );


	//  outer loop
	for ( i=0u; i < maxTemperatureSteps; i++ )
	{
		//  select random partner
		iPartner  =  getRandomPartner( self );
		copySolution( pPsa, postedSolution+iPartner );

		//  inner loop
		for ( j=0u; j < maxIterations; j++ )
		{
			sa->twoOpt( pPsa, common );
			deltaE  =  double(sa->getImprovement());

			if ( deltaE >= 0. )
				sa->update();
			else
				if (Random::random() < exp(deltaE/currentTemp))
					sa->update();
		}
		//  end of inner loop

		sa->backToBest(true);

		if ( sa->getCurrentLength() < bestLength )
		{
			currentTemp  =  currentTemp * mu;
			bestLength  =  sa->getCurrentLength();
		}


		printf( "%3d:%3d", self, i );
		printf( ",   currentTemp[%1d] = %g", 0, currentTemp );
		printf( ",   %10d\n", sa->getCurrentLength());


		postSolution( self, sa );

	}
	//  end of outer loop


	// reporting

	sLength[self]  =  sa->getCurrentLength();



	delete  sa;			// good housekeeping

	return  ! SetEvent( hEvent3[self] );
}





void  copySolution( Tour* to, Tour* from )
{
	EnterCriticalSection( &cs );
	to->copyTour( from );
	LeaveCriticalSection( &cs );
}




unsigned  getRandomPartner( const unsigned& butNotSelf )
{
	unsigned  retVal = unsigned(Random::random()*nThreads);

	while ( retVal == butNotSelf )
		retVal  =  unsigned(Random::random()*nThreads);

	return  retVal;
}




unsigned  getSolLength( const unsigned& thrIndex )
{
	return  postedSolution[thrIndex].getCurrentLength();
}




void  postSolution( const unsigned& self, Tour* sa )
{
	copySolution( postedSolution+self, sa );
}
